/*
    compile with: grep ^gcc bits2bitmap.c  | sh
gcc -o bits2bitmap bits2bitmap.c
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <netinet/in.h>

/* most .wav files: header size 44 */

#define IMGWIDTH    1024        // has to be divisible by 8
#define IMGHEIGHT    1024

void *extract_bits(const char *read, unsigned bit, unsigned size, unsigned stride);
char *make_imgname(const char *infname, unsigned skip, unsigned bit, unsigned numchannels, unsigned channel);
void *load_file(const char *filename, long *filesize);
int write_pbm( const char *fname, char *buf, int width, int height );

int main(int argc, char **argv)
{
    char *buf, *read, *image, *outfname;
    long size;
    unsigned skip, bitnr, stride, channels, ch, nextskip;

    if( argc < 4 || argc > 5 || !isdigit(*argv[2]) || !isdigit(*argv[3]) ||
                                        (argc == 5 && !isdigit(*argv[4])) ) {
        printf("usage: bits2bitmap <file> <skip>[.<bit>] <stride> [<channels>]\n");
        return 1;
    }
    buf= load_file(argv[1], &size);
    if( !buf )
        return 2;
    skip= strtoul(argv[2], &read, 0);
    if( *read == '.' && isdigit(read[1]) )
        bitnr= strtoul(read+1, NULL, 0);
    else
        bitnr= 0;
    stride= strtoul(argv[3], NULL, 0);
    channels= argc == 5 ? strtoul(argv[4], NULL, 0) : 1;
    size -= skip;
    for( ch= 0; ch< channels; ++ch ) {
        image= extract_bits(buf+skip+ch*stride/channels, bitnr, size, stride);
        outfname= make_imgname(argv[1], skip, bitnr, channels, ch);
        write_pbm(outfname, image, IMGWIDTH, IMGHEIGHT);
        free(outfname);
        free(image);
    }
    free(buf);
    skip += IMGWIDTH * IMGHEIGHT * stride;
    size -= IMGWIDTH * IMGHEIGHT * stride;
    if( size > 0 ) {
        size -= IMGWIDTH * IMGHEIGHT * stride;
        printf("Skip value for next set of data: %d == 0x%X  %s\n",
                skip, skip, size > 0? "" : "(incomplete)");
    }
    return 0;
}


void *extract_bits(const char *read, unsigned bit, unsigned size, unsigned stride)
{
    unsigned char *img, *write;
    unsigned i, j, bitnr, pixval;
    unsigned char byte= 0;

    img= malloc(IMGWIDTH * IMGHEIGHT / 8);
    memset(img, 0xFF, IMGWIDTH * IMGHEIGHT / 8);
    if( size > IMGWIDTH * IMGHEIGHT * stride )
        size= IMGWIDTH * IMGHEIGHT * stride;
    size -= 7*stride;
    write= img;
    for( i= 0; i < size; ) {
        for( j= 8; j > 0; --j, i += stride )
            byte= (byte << 1) | ((read[i] >> bit) & 1);
        *write++ = byte;
    }
    return img;
}


#define SUFFIXLEN   (6 + 11 + 6 + 5)

char *make_imgname(const char *infname, unsigned skip, unsigned bit, unsigned numchannels, unsigned channel)
{
    char *imgname, *scan;
    unsigned iflength;

    iflength= strlen(infname);
    imgname= malloc(iflength + SUFFIXLEN);
    strcpy(imgname, infname);
    for( scan= imgname + iflength; scan > imgname && *scan != '.' && *scan != '/'; --scan );
    if( scan == imgname || *scan == '/' )
        scan += iflength;
    if( numchannels == 1 )
        snprintf(scan, SUFFIXLEN, ".bit%d-%X.ppm", bit, skip);
    else
        snprintf(scan, SUFFIXLEN, ".bit%d-%X-ch%d.ppm", bit, skip, channel);
    return imgname;
}


/*!
  \brief Read a file into RAM

  Allocates RAM and reads all of a file into it.  A zero byte is appended
  (useful for text files), and the size is written to the second argument.
  If unsuccessful, file error messages go into stderr.

  \param filename File name
  \param filesize Pointer to long variable for file/buffer size (or NULL)

  \return Pointer to file data, or NULL if the operation failed.
*/
void *load_file(const char *filename, long *filesize)
{
    FILE *in;
    void *buf;
    long size;

    in= fopen(filename, "r");
    if( !in ) {
        fprintf(stderr, "Could not open `%s' for reading.\n", filename);
        return NULL;
    }
    fseek(in, 0, SEEK_END);
    size= ftell(in);
    fseek(in, 0, SEEK_SET);
    buf= malloc(size+1);
    if( !buf ) {
        fprintf(stderr, "Could not allocate memory for loading `%s'.\n", filename);
        fclose(in);
        return NULL;
    }
    if( ferror(in) || fread(buf, 1, size, in) != size ) {
        fprintf(stderr, "Error reading `%s'.\n", filename);
        free(buf);
        fclose(in);
        return NULL;
    }
    fclose(in);
    ((char*)buf)[size]= 0;
    if( filesize )
        *filesize= size;
    return buf;
}


/*---------------------------------------------------------------------------
  write_pbm	Write bitmap image to PBM file

  ->	fname	file name
	buf	image data in row-major, format; MSB of each byte is the leftmost pixel
	width	image width (has to be divisible by 8)
	height	image height
  <-  0 success; 1 error
---------------------------------------------------------------------------*/

int write_pbm( const char *fname, char *buf, int width, int height )
{
    FILE *out;

    out= fopen(fname, "wb");
    if( !out ) {
        fprintf(stderr, "Could not create output file `%s'.  Aborting.\n", fname);
        return 1;
    }
    fprintf(out, "P4\n%d %d\n", width, height);
    fwrite(buf, 1, width*height/8, out);
    if( ferror(out) ) {
        fprintf(stderr, "Error writing `%s'.\n", fname);
        fclose(out);
        return 1;
    }
    fclose(out);
    return 0;
}


