/*
    compile with: grep ^gcc bits2points.c  | sh
gcc -o bits2points bits2points.c
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <netinet/in.h>

/* most .wav files: header size 44 */

#define COORDBITS   10
#define DENSITY     64         // inverse density, really

void *extract_points(const char *read, unsigned bit, unsigned size, unsigned stride);
char *make_imgname(const char *infname, unsigned skip, unsigned bit, unsigned numchannels, unsigned channel);
void *load_file(const char *filename, long *filesize);
int write_pgm( const char *fname, char *buf, int width, int height, int maxval );

int main(int argc, char **argv)
{
    char *buf, *read, *image, *outfname;
    long size;
    unsigned skip, bitnr, stride, channels, ch, nextskip;

    if( argc < 4 || argc > 5 || !isdigit(*argv[2]) || !isdigit(*argv[3]) ||
                                        (argc == 5 && !isdigit(*argv[4])) ) {
        printf("usage: bits2points <file> <skip>[.<bit>] <stride> [<channels>]\n");
        return 1;
    }
    buf= load_file(argv[1], &size);
    if( !buf )
        return 2;
    skip= strtoul(argv[2], &read, 0);
    if( *read == '.' && isdigit(read[1]) )
        bitnr= strtoul(read+1, NULL, 0);
    else
        bitnr= 0;
    stride= strtoul(argv[3], NULL, 0);
    channels= argc == 5 ? strtoul(argv[4], NULL, 0) : 1;
    size -= skip;
    for( ch= 0; ch< channels; ++ch ) {
        image= extract_points(buf+skip+ch*stride/channels, bitnr, size, stride);
        outfname= make_imgname(argv[1], skip, bitnr, channels, ch);
        write_pgm(outfname, image, (1 << COORDBITS), (1 << COORDBITS), 255);
        free(outfname);
        free(image);
    }
    free(buf);
    skip += (1 << (2*COORDBITS)) / DENSITY * 2*COORDBITS * stride;
    size -= (1 << (2*COORDBITS)) / DENSITY * 2*COORDBITS * stride;
    if( size > 0 ) {
        size -= (1 << (2*COORDBITS)) / DENSITY * 2*COORDBITS * stride;
        printf("Skip value for next set of data: %d == 0x%X  %s\n",
                skip, skip, size > 0? "" : "(incomplete)");
    }
    return 0;
}


void *extract_points(const char *read, unsigned bit, unsigned size, unsigned stride)
{
    unsigned char *img;
    unsigned long long pixel;
    unsigned mask, count, bitnr, pixval;

    img= malloc(1 << (2*COORDBITS));
    memset(img, 0xFF, 1 << (2*COORDBITS));
    mask= 1 << bit;
    count= (1 << (2*COORDBITS)) / DENSITY;
    if( count*2*COORDBITS*stride > size )
        count= size/(2*COORDBITS*stride);
    for( ; count > 0; --count ) {
        pixel= 0;
        for( bitnr= 2*COORDBITS; bitnr > 0; --bitnr ) {
            pixel <<= 1;
            pixel += mask & *(const unsigned char*)read;
            read += stride;
        }
        pixel >>= bit;
        if( img[pixel] )
            img[pixel] -= 255 / 3;
    }
    return img;
}


#define SUFFIXLEN   (5 + 11 + 6 + 5)

char *make_imgname(const char *infname, unsigned skip, unsigned bit, unsigned numchannels, unsigned channel)
{
    char *imgname, *scan;
    unsigned iflength;

    iflength= strlen(infname);
    imgname= malloc(iflength + SUFFIXLEN);
    strcpy(imgname, infname);
    for( scan= imgname + iflength; scan > imgname && *scan != '.' && *scan != '/'; --scan );
    if( scan == imgname || *scan == '/' )
        scan += iflength;
    if( numchannels == 1 )
        snprintf(scan, SUFFIXLEN, ".bit%d-%X.pgm", bit, skip);
    else
        snprintf(scan, SUFFIXLEN, ".bit%d-%X-ch%d.pgm", bit, skip, channel);
    return imgname;
}


/*!
  \brief Read a file into RAM

  Allocates RAM and reads all of a file into it.  A zero byte is appended
  (useful for text files), and the size is written to the second argument.
  If unsuccessful, file error messages go into stderr.

  \param filename File name
  \param filesize Pointer to long variable for file/buffer size (or NULL)

  \return Pointer to file data, or NULL if the operation failed.
*/
void *load_file(const char *filename, long *filesize)
{
    FILE *in;
    void *buf;
    long size;

    in= fopen(filename, "r");
    if( !in ) {
        fprintf(stderr, "Could not open `%s' for reading.\n", filename);
        return NULL;
    }
    fseek(in, 0, SEEK_END);
    size= ftell(in);
    fseek(in, 0, SEEK_SET);
    buf= malloc(size+1);
    if( !buf ) {
        fprintf(stderr, "Could not allocate memory for loading `%s'.\n", filename);
        fclose(in);
        return NULL;
    }
    if( ferror(in) || fread(buf, 1, size, in) != size ) {
        fprintf(stderr, "Error reading `%s'.\n", filename);
        free(buf);
        fclose(in);
        return NULL;
    }
    fclose(in);
    ((char*)buf)[size]= 0;
    if( filesize )
        *filesize= size;
    return buf;
}


/*---------------------------------------------------------------------------
  write_pgm	Write greyscale image to PGM file

  ->	fname	file name
	buf	image data in row-major, host-endian format; will be destroyed
		for >8-bit images by endian conversion
	width	image width
	height	image height
	maxval	maximum pixel value
  <-  0 success; 1 error
---------------------------------------------------------------------------*/

int write_pgm( const char *fname, char *buf, int width, int height, int maxval )
{
    FILE *out;
    int count, size, pixbytes;

    size= width*height;
    pixbytes= maxval > 255 ? 2 : 1;
    if( maxval > 255 )
        for( count= 0; count< size; ++count )
            ((short*)buf)[count]= htons(((short*)buf)[count]);
    out= fopen(fname, "wb");
    if( !out ) {
        fprintf(stderr, "Could not create output file `%s'.  Aborting.\n", fname);
        return 1;
    }
    fprintf(out, "P5\n%d %d\n%d\n", width, height, maxval);
    fwrite(buf, pixbytes, size, out);
    if( ferror(out) ) {
        fprintf(stderr, "Error writing `%s'.\n", fname);
        fclose(out);
        return 1;
    }
    fclose(out);
    return 0;
}


