/*
    compile with: grep ^gcc getbit.c  | sh
gcc -o getbit getbit.c
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

/* most .wav files: header size 44 */

char *getoutfname(const char *infilename, unsigned numchannels, unsigned channel, unsigned bit);
void writebits(FILE *out, char *read, unsigned bit, unsigned size, unsigned stride);
void *load_file(const char *filename, long *filesize);

int main(int argc, char **argv)
{
    FILE *outfile;
    char *buf, *read, *outfname;
    long size;
    unsigned skip, bitnr, stride, channels, ch;

    if( argc < 4 || argc > 5 || !isdigit(*argv[2]) || !isdigit(*argv[3]) ||
                                        (argc == 5 && !isdigit(*argv[4])) ) {
        printf("usage: getbit <file> <skip>[.<bit>] <stride> [<channels>]\n");
        return 1;
    }
    buf= load_file(argv[1], &size);
    if( !buf )
        return 2;
    skip= strtoul(argv[2], &read, 0);
    if( *read == '.' && isdigit(read[1]) )
        bitnr= strtoul(read+1, NULL, 0);
    else
        bitnr= 0;
    stride= strtoul(argv[3], NULL, 0);
    channels= argc == 5 ? strtoul(argv[4], NULL, 0) : 1;
    size -= skip;
    for( ch= 0; ch< channels; ++ch ) {
        outfname= getoutfname(argv[1], channels, ch, bitnr);
        if( !outfname ) {
            fprintf(stderr, "Could not allocate space for output file name.\n");
            continue;
        }
        outfile= fopen(outfname, "wb");
        if( !outfile ) {
            perror("Error opening output file");
            fprintf(stderr, "Could not create %s\n", outfname);
            continue;
        }
        writebits(outfile, buf+skip+ch*stride/channels, bitnr, size, stride);
        if( ferror(outfile) ) {
            perror("");
            fprintf(stderr, "Error writing %s.\n", outfname);
        }
        if( ferror(outfile) )
            perror("Error writing output data");
        fclose(outfile);
        free(outfname);
    }
    free(buf);
}


#define SUFFIXLEN   11

char *getoutfname(const char *infilename, unsigned numchannels, unsigned channel, unsigned bit)
{
    char *name;
    int origlen, extpos, len;

    origlen= strlen(infilename);
    len= origlen + SUFFIXLEN;
    name= malloc(len);
    if( !name )
        return NULL;
    for( extpos= origlen - 1; extpos > 0 && infilename[extpos] != '/' &&
                              infilename[extpos] != '.'; --extpos );
    if( extpos == 0 || infilename[extpos] == '/' )
        extpos= origlen;
    strncpy(name, infilename, extpos);
    if( numchannels == 1 )
        snprintf(name+extpos, SUFFIXLEN, ".bit%d", bit);
    else
        snprintf(name+extpos, SUFFIXLEN, ".ch%dbit%d", channel, bit);
    return name;
}


void writebits(FILE *out, char *read, unsigned bit, unsigned size, unsigned stride)
{
    unsigned count[2];
    unsigned i, j;
    unsigned char byte= 0;

    size -= 7*stride;
    for( i= 0; i < size; ) {
        for( j= 8; j > 0; --j, i += stride )
            byte= (byte << 1) | ((read[i] >> bit) & 1);
        fwrite(&byte, 1, 1, out);
    }
}


/*!
  \brief Read a file into RAM

  Allocates RAM and reads all of a file into it.  A zero byte is appended
  (useful for text files), and the size is written to the second argument.
  If unsuccessful, file error messages go into stderr.

  \param filename File name
  \param filesize Pointer to long variable for file/buffer size (or NULL)

  \return Pointer to file data, or NULL if the operation failed.
 */
void *load_file(const char *filename, long *filesize)
{
    FILE *in;
    void *buf;
    long size;

    in= fopen(filename, "r");
    if( !in ) {
        fprintf(stderr, "Could not open `%s' for reading.\n", filename);
        return NULL;
    }
    fseek(in, 0, SEEK_END);
    size= ftell(in);
    fseek(in, 0, SEEK_SET);
    buf= malloc(size+1);
    if( !buf ) {
        fprintf(stderr, "Could not allocate memory for loading `%s'.\n", filename);
        fclose(in);
        return NULL;
    }
    if( ferror(in) || fread(buf, 1, size, in) != size ) {
        fprintf(stderr, "Error reading `%s'.\n", filename);
        free(buf);
        fclose(in);
        return NULL;
    }
    fclose(in);
    ((char*)buf)[size]= 0;
    if( filesize )
        *filesize= size;
    return buf;
}


